/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

//  xxxxxxxxxxxxx          version: 1.0              xxxxxxxxxxxxxxx
		 
//  xxxxxxxxxxxxxxx   global variable declarations   xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var ptype = '1';  // Paradigm type: '1' => 100% partial report. or: '2' => 50% partial report

var item_cnt = 4;    // The number of stimuli in the task

var alfa_type = '1';  // Type of Alphanumeric characters => "1" = Digits; "2" = Letters-lowercase; "3" = Letters-UPPERCASE; "4" = All digits and letters

var alfa_cnt = 3;  // Number of of alphanumeric characters in each row

var ttype = '1';   // Task type (and score type). range= '1'..'3'. see html ttype element

var font_family = 'Georgia, serif';  // default canvas ctx.fontFamily 

var arrows = '2';  // Arrows position. '1' => Left, '2' => Right, '3' => Left and right

var rep_cnt = 1;       // Maximum number of stimulus repetitions within the same item

var before = 1500;       // Delay time before show each stimulus (ms).

var s_time = 10;     //     Alphanumeric characters duration (ms) => Duration of main stimulus (digits or letters)

var ad_time = 100;    // Arrow delay time (ms)

var a_time = 10;  // Arrow duration time (ms)

var after = 1000;				 // Delay time after show each stimulus (ms)

var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var is_mis = true;    // if (is_mis==true) => missing values are allowed

var prg_type = true;   //  type of information in Progress Bar (true => Items, false => Nothing) 

var demo = false;     // if (demo == true) => Show task solutions 
  
var back_clr = '#fff';  // background canvas color during task

var pen_clr = '#000';  //  font color (base font) during the task 
	
var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 
 

// Other global constants and variables:

var is_storage = false; // local storage is enabled on user browser or not
   
var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_task = false;    // task in progress or not. is_task is true for all events between startTask() and theEnd() events

var startTime;      // timestamp (start time) for current stimulus

var item = 0;   		// current item in task     

var counter = 0;     // toogle counter (0 or 1) for window.requestAnimationFrame(draw) callback and window.requestAnimationFrame(getFPS) callback

var FPS = 0;     // frames per second. Measured Frame rate on user screen. see: function getFPS() 

var repeat_count = 0;    // number of repeated items in whole task

// global (window) arrays
	
var stim = [];   // stimulus length for each item. range dimensions

var ans = [];   //  the user's response to each item in the task

var repeated = [];  // for each item: the number of completed repetitions of the sound stimulus for each item. 

var t_row = [];     //  For each item in the task: which row is the target (0..2), ie which row should be remembered 

var is_arrow = [];  //  For each item in the task: if true => this item have arrows in delay period (else there is no arrows until end)



// application init [check storage, show start panel..] 
function init() 
{   
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }			
	let e= document.getElementById('answer');  // attach event handlers..
	e.onkeypress = answerKeyPress;	
	e.onkeyup = answerKeyUp;	
	//e.onpaste= function(){return false}; // Prevent pasting in answer input type=text (since pasted content might include non-allowed characters)
		
	// do FPS calibration..	
	setCanvas(document.getElementById('can_s'));	
	window.requestAnimationFrame(getFPS);  
	
	loadFont();
	setIns();
}


/* https://www.viewsonic.com/library/tech/frame-rate-vs-refresh-rate-whats-the-difference/
	 https://en.wikipedia.org/wiki/Frame_rate
	Frame rate is produced by a combination of your graphics card and your processor, so it is essentially the number of frames a system is able to produce in a second, whereas refresh rate is the rate at which a monitor is able to completely refresh its display.  */
function getFPS(timestamp) 
{ 
	if (counter == 0) {  
		startTime = timestamp;	// Remember the time when the stimulus display started 
		FPS= 0;  // probably redundant
		document.getElementById('can_s').style.display= 'block';  	  	
	}	
	counter= 1;	
	let elapsed = timestamp - startTime;  // const elapsed = Math.round(timestamp - start); 						
	if ( elapsed <= 1000  ) {    // just wait
		FPS++;
		window.requestAnimationFrame(getFPS);  	// Note: Your callback routine must itself call requestAnimationFrame() again if you want to animate another frame at the next repaint. requestAnimationFrame() is 1 shot.				
	}	
	else  {  // hide stimulus 	
	  let freq= (elapsed/FPS).toFixed(3);
	  document.getElementById('fps').textContent= ' Measured Frame rate (frames per second or FPS or frame frequency) FPS = ' + FPS + ' Hz. Tested time period = 1000 ms.';
    document.getElementById('fps_2').textContent=  'Theoretical Minimum duration of stimuli on this device = ' + freq + 'ms. Maximum reliable step value to increase stimulus length on this device = ' + freq + 'ms.'		
		document.getElementById('can_s').style.display= 'none';	
		showStart();   		
		//alert('Calibration done. \n Measured Frame rate (FPS, frames per second, frame frequency) FPS = ' + FPS + ' Hz. \n Tested time period = 1000 ms. \n  Theoretical Minimum duration of stimuli on this device = ' + freq + ' ms. \n  Maximum reliable step value to increase stimulus length on this device = ' + freq + ' ms.');
	}			
}


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex=1; // Georgia, serif
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[1];
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		   
 
	if ( alfa_type == '1' )  {
		document.getElementById('ins_alfa_type').textContent= 'digits';
	}
	else if ( alfa_type == '4' ) {
	  document.getElementById('ins_alfa_type').textContent= 'digits and letters';
	}
  else  {
		document.getElementById('ins_alfa_type').textContent= 'letters';  // '2' or '3'
	}
		
	document.getElementById('ins_ptype').style.display = ptype == '2' ? 'block' : 'none';
	  
  if ( ttype == '1' )  {
		document.getElementById('ins_ttype').textContent= 'You do not have to remember the position of the character within the row, ie the order in which the required characters are entered is irrelevant.';
	}
	else if (ttype == '2') {
	  document.getElementById('ins_ttype').textContent= 'You must remember each character within the required row but also the position of each character within the required row in the LEFT-RIGHT direction.';
	}
  else if (ttype == '3') {
		document.getElementById('ins_ttype').textContent= 'You must remember each character within the required row but also the position of each character within the required row in the RIGHT-LEFT direction.';
	}
		
	if (rep_cnt > 1) {		
		document.getElementById('ins_rep_cnt').style.display = 'block';
		document.getElementById('ins_rep_cnt').textContent= 'If you want to display the new stimulus in the current item, click the "Repeat" button and the new stimulus will be displayed. You can use this option ' + rep_cnt + '  times within the same item.';
	}
	else  {
		document.getElementById('ins_rep_cnt').style.display = 'none';
	}
	
  document.getElementById('ins_mis').style.display = is_mis ? 'none' : 'block';  
	document.getElementById('ins_demo').style.display = demo ? 'block' : 'none';
	
}	


// raised when page is resized. see html: <body onresize="pageResize()"> => stretch canvas 
function pageResize()  
{   
	if (is_task)  { 
		theEnd(true);	  // resize error										 
	}	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{

		case 'ptype':                                                                                                                                           
			msg.textContent= e.value == '1' ? 'All stimuli are displayed with an arrow in the delay period.' : '50% of stimuli are displayed with an arrow in the delay period. For the other 50% of stimuli, this information is available only at the end, ie when entering alphanumeric characters.';
			break;

		case 'item_cnt':		 
		case 's_time':  		
		case 'a_time': 
		case 'after': 		
			if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 			
			break;  
		
		case 'before':
			if (e.validity.valid)  {
				msg.textContent= 'Stimuli are prepared in memory during this period. So leave this period longer, especially for very short stimulus displays.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 			
      break;
			
		case 'rep_cnt':
			if (e.validity.valid)  {
				msg.textContent= 'If you set this setting greater than 1, when the user clicks the "Repeat" button, a NEW stimulus will be generated for the same item. It can be useful, for very short stimuli that the user has failed to notice.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 		 
		  break;
		
		case 'ad_time': 
		if (e.validity.valid)  {
				msg.textContent= 'The time between the disappearance of the alphanumeric characters and the start of the arrow display.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
      document.getElementById('ad_time_0').style.display= e.value == 0 ? 'block': 'none';			
			break;
				
    case 'alfa_cnt': 
			if (e.validity.valid)  {
				msg.textContent= 'For digits the allowed range is 1..6. Otherwise: 1..18'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;
		
		//case 'ttype':
		  //break;
		
		//case 'font_family':
		  //break;
		
		case 'arrows':
		  if (e.value == '1')  {
				msg.textContent= 'Arrows indicating which row to remember appear on the left side of the screen.';
			}
			else  if (e.value == '2')  {
				msg.textContent= 'Arrows indicating which row to remember appear on the right side of the screen.';
			}
			else  {
				msg.textContent= 'Arrows that indicate which row to remember appear at the same time on the left and right sides of the screen.';
			}
		  break;
				
    case 'alfa_type':	 // Type of Alphanumeric characters => "1" = Digits; "2" = Letters-lowercase; "3" = Letters-UPPERCASE; "4" = All digits and letters	
			document.getElementById('alfa_cnt').max= (e.value == '1') ? 6 : 18; 				
			if (document.getElementById('alfa_cnt').validity.valid)  {
				document.getElementById('alfa_cnt_msg').textContent= 'For digits the allowed range is 1..6. Otherwise: 1..18.'; 
				document.getElementById('alfa_cnt_msg').removeAttribute('class');
			}
			else  {
				document.getElementById('alfa_cnt_msg').textContent= s + document.getElementById('alfa_cnt').min + ".." + document.getElementById('alfa_cnt').max;
				document.getElementById('alfa_cnt_msg').setAttribute('class', 'error');
			}
			break;  			
		
    case 'is_back':  		
			msg.textContent= e.checked ? 'Yes' : 'No. Returning to previous items is not allowed.';    
			break;
    
    case 'is_mis': 
    	msg.textContent= e.checked ? 'Missing values are allowed. The user does not have to enter a value.' : 'Missing values are not allowed. The user must enter any non-empty value.';    
			break;
			    
    case 'prg_type':                                                                                                                                                   
      msg.textContent= e.checked ? 'Progress bar will show the number of the current item and the number of remaining items until the end of the task.' : 'Progress bar will be invisible';     
      break;    
				
		case 'demo':
			msg.textContent= e.checked ? 'The correct solution will be displayed for each item.' : 'No';  
			break; 
		
		case 'back_clr': 		// background canvas color during task
		case 'pen_clr':  		//  font color (base font) during the task   
			msg.textContent= ( validColor(e.value) ) ? '' : 'The selected color is invalid.'; 
			break; 	
		
		case 'store':                                                                                                                                           
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty == true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number" + check colors
function optionsValid()  
{      
	let res= document.getElementById('item_cnt').validity.valid && 
	document.getElementById('alfa_cnt').validity.valid && 	
	document.getElementById('s_time').validity.valid && 
	document.getElementById('ad_time').validity.valid && 
	document.getElementById('a_time').validity.valid && 
	document.getElementById('before').validity.valid &&	
	document.getElementById('after').validity.valid &&	
	document.getElementById('rep_cnt').validity.valid &&
	validColor(document.getElementById('back_clr').value) && 
	validColor(document.getElementById('pen_clr').value);
  return res;              
}


// start task (research) 
function startTask(e)   
{     
  initLayout();  

	setPar(); 
	
	stimGen();
 
  showItem();     
}


// init page layout when task is started. init visible/invisible html dom elements, progress bar
function initLayout()
{

	setCanvas(document.getElementById('can_s'));	
	setCanvas(document.getElementById('can_a'));	
	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
  document.getElementById('task_section').style.display= 'block'; 
	document.getElementById('item_div').style.display= 'none';
	
	document.getElementById('nav_items').textContent= '';	
  if (!prg_type)  {
		document.getElementById('progress_holder').style.display= 'none';    	
	}	
	else  {
		document.getElementById('progress_holder').style.display= 'block';   				   			
		document.getElementById('progress').setAttribute('style', 'width:100%');  // reset progress bar width						 	
	}
  
	document.getElementById('demo_sol').style.display= demo ? 'block': 'none';
	
	// see: wm14 - v1.1
	switch (alfa_type)
	{		
		case '2':  // Letters-lowercase 
			document.getElementById('answer').setAttribute('style','text-transform:lowercase');			
			break;
		case '3':  // Letters-UPPERCASE 
			document.getElementById('answer').setAttribute('style','text-transform:uppercase');			
			break;
		default:  // Digits ("1") or All digits and letters ("4")
			document.getElementById('answer').removeAttribute('style');	
	} 
	
	document.getElementById('answer').setAttribute('maxLength', alfa_cnt);	
	
	/* ttype => Task type (and score type)
	if (ttype == '1')  {		
		document.getElementById('answer').setAttribute('maxLength', alfa_cnt);			
	}
	else  { 
		document.getElementById('answer').removeAttribute('maxLength');		
	} */
}


// center stimulus (canvas) on user screen (approx)
// c => canvas element => document.getElementById('can_s') or document.getElementById('can_a')
function setCanvas(c)  
{
	let ctx= c.getContext('2d');	 
	c.width= innerWidth;  
	c.style.width= c.width + 'px';	     
	c.height= innerHeight;   
	c.style.height= c.height + 'px';  

  ctx.clearRect(0, 0, c.width, c.height);  
	c.style.display= 'block';
	
	ctx.textAlign = 'center';
	ctx.textBaseline= 'middle';	
	
	ctx.fillStyle= back_clr;
	ctx.strokeStyle= pen_clr; 	
	
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp
	ctx.font= '10px Arial, Verdana, Helvetica, sans-serif';  	
}


// set internal application research parameters
// reset all global vars and arrays
function setPar()  
{
	//if (item_cnt % 2 != 0) item_cnt++;
  if ( alfa_type == '1' && alfa_cnt > 6 )  {  // Type of alphanumeric characters: 'Digits' and Number of of alphanumeric characters in each row > 6
		alfa_cnt= 6;  // digits max = 6 in each row
	}
	
	// init all global arrays
	stim = []; 
	ans= []; 	
	repeated= [];
	t_row = []; 
	is_arrow = [];
	
	for (let i= 0; i< item_cnt; i++)   {  // i => item        
    ans[i]= '';	
		repeated[i]= 0;
		is_arrow[i]= true;
  } 
	
	// reset all global vars
  item= 0;  	
	repeat_count= 0;
	is_task= true;     //  see pageResize()
	counter = 0;	
}


// stimulus generator. generates task stimuli 
function stimGen()	
{
	// generate string content	
	for (let i= 0; i < item_cnt; i++)  {		
		newItem(i);
	}
	
	// determine which row is the stimulus (0..2) should be memorized 
	for (let item= 0; item < item_cnt; item++)   {
    t_row[item]= getRndInt(0,2);
  }		
	
  // For ptype == '2' => determine which stimulus (item) will NOT have arrows in delay period. see: setPar()
	if (ptype == '2')  {
		let sel= {}, j;	
		for (let i = 1; i <= item_cnt/2; i++)  
		{  // 50% will be without arrows in delay period
			do  {	
				j= getRndInt(0,item_cnt-1);
			} while (  typeof sel[j] !== 'undefined' );					  
			sel[j]= 1;  // save this item index
			is_arrow[j]= false;
		}		
	}
}


/* generate new item. i => item number. range= [0..item_cnt-1]
   each character in each row must be different
   if a new stimulus is generated for the same item (rep_cnt > 1), 
	 this new stimulus will be different from the previous one in all 3 rows */ 
function newItem(i) 
{
	let prev= stim[i], ok;  // prev=> previous stimulus (if exists)
	stim[i]= [];  //	 1. index => item.	2. index 0..2 => row
	do 
	{ 
	  ok= true;
		for (let row = 0; row < 3; row++)
		{
			let s= '';  // characters in each row (0..2)
			switch (alfa_type)
			{
				case '1':  // digits 		
					let key= {}, d;
					for (let i=0; i<alfa_cnt; i++)  {
						do  {
							d= getRndInt(0,9).toString();
						}	while ( key[d] == 1 );   // each character in each row must be different
						key[d] = 1;
						s+= d;
					}			  
					break;			
				case '2':  // Letters-lowercase 
					s= lowerStrGen(alfa_cnt);			
					break;
				case '3':  // Letters-UPPERCASE 
					s= upperStrGen(alfa_cnt);			
					break;
				default:  // 4 => All digits and letters 
					s= allStrGen(alfa_cnt);
			}
			if (prev) 
			{ // if exists previous: new stimulus will be different from the previous one in all 3 rows
				for (let row = 0; row < 3; row++)  
				{
					if (prev[row] == s)  {
						ok= false;
						break;
					}
				}
			}			
			stim[i][row]= s;
		}
	} 
	while ( !ok );
	
}


// generates a random string from lowercase letters
// each generated character must be different
function lowerStrGen(cnt)
{ 
  let all = "abcdefghijklmnopqrstuvwxyz";
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
		do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
    txt += c;
	}
  return txt;
}


// generates a random string from uppercase letters
// each generated character must be different 
function upperStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
    do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
		txt += c;
	}
  return txt;
}


// generates a random string from all (uppercase and lowercase) letters and digits
// each generated character must be different 
function allStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
    do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
		txt += c;
	}
  return txt;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	if ( document.getElementById('ptype').selectedIndex > - 1 )  {
		ptype= document.getElementById('ptype').value;
	}
  
  if ( document.getElementById('item_cnt').validity.valid )  {  
		item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  }

  if ( document.getElementById('alfa_type').selectedIndex > - 1 )  {
		alfa_type= document.getElementById('alfa_type').value;
	}

  if ( document.getElementById('alfa_cnt').validity.valid )  {  
		alfa_cnt= parseInt(document.getElementById('alfa_cnt').value, 10);			
  } 

  if ( document.getElementById('ttype').selectedIndex > - 1 )  {
		ttype= document.getElementById('ttype').value;
	}

	font_family= document.getElementById('font_family').value;

  if ( document.getElementById('arrows').selectedIndex > - 1 )  {
		arrows= document.getElementById('arrows').value;
	}

  if (document.getElementById('rep_cnt').validity.valid) {
		rep_cnt= parseInt(document.getElementById('rep_cnt').value, 10);	
  }

  if (document.getElementById('before').validity.valid) {
		before= parseInt(document.getElementById('before').value, 10);
  }

	if ( document.getElementById('s_time').validity.valid )  {  
		s_time= parseInt(document.getElementById('s_time').value, 10);			
  }
	 	
	if ( document.getElementById('ad_time').validity.valid ) {  
		ad_time= parseInt(document.getElementById('ad_time').value, 10);			
  }  
	 
	if (document.getElementById('a_time').validity.valid ) {
		a_time= parseInt(document.getElementById('a_time').value, 10);	
  } 
	  
	if (document.getElementById('after').validity.valid) {
		after= parseInt(document.getElementById('after').value, 10);  
  }

	is_back= document.getElementById('is_back').checked ? true: false;       
  
  is_mis= document.getElementById('is_mis').checked ? true: false; 
	
	prg_type= document.getElementById('prg_type').checked ? true : false;  // display item count on footer or nothing
		
	demo= document.getElementById('demo').checked ? true: false;	
		
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr').value))  {		
		pen_clr= document.getElementById('pen_clr').value;
  }
		
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}

// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{  
  ans[item]= trim(document.getElementById('answer').value);

	if (item > 0)  {
    item--;  
    showItem();     // show again Panel for previous item 
  }         
}


// goto next item (stimulus)
function nextItem()  
{ 	
	if ( !valAnswer() )  {  
		document.getElementById('req_val').style.display= 'block';    //alert('You must enter any non-blank answer. Try to guess.'); 
		return;
  }	
  
	ans[item]= trim(document.getElementById('answer').value);	
	item++;					
	if (item < item_cnt)  {         		 
    showItem();     // show next item        
  }
  else  {               
    theEnd(false);  // end of task. no resize error in this task 
  }
}


// trim input string (user answer)
function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}
 
 
/* see function init().  deprecated but still works:
   https://developer.mozilla.org/en-US/docs/Web/API/GlobalEventHandlers/onkeypress
   this regex [1-9()*\/+-] was tested here:    https://regex101.com/    */
function answerKeyPress(event)   { 
  if (alfa_type == '1') {   // only digits
		return event.charCode === 0 || /[0-9]/.test(String.fromCharCode(event.charCode));	
	}
	else {
		return true;
	}
}


// go to the next task if the ENTER key is pressed.
function answerKeyUp(e)  
{  
	if ( !is_mis && trim(document.getElementById('answer').value) == '' )  {
		document.getElementById('req_val').style.display=  'block'; 
	}
	else  {
		document.getElementById('req_val').style.display= 'none'; 
	}
	if ( e.key === 'Enter' || e.keyCode === 13 ) {  
		nextItem();
	}
}


// validate user answer. par: v => user answer value for current item
function valAnswer()  
{	
	if ( is_mis ) {   // missing values are allowed
	  return true;
	}
	else {   // missing values are not allowed
    let res = ( trim(document.getElementById('answer').value) == '' ) ? false : true;		
		return res; 
  }
}


// handle alphanumeric characters duration  
// trigger end of display of alphanumeric characters and begin of arrow  
function displayString(timestamp)  {
	let elapsed = timestamp - startTime; 
	if ( elapsed < s_time ) {  //   var s_time => alphanumeric characters duration (ms)
		window.requestAnimationFrame(displayString);  // wait
	}
	else  {   // var s_time => Alphanumeric characters duration (ms)  	
		document.getElementById('can_s').style.visibility = 'hidden';  // only once  
		startTime = timestamp;	// timestamp when the alphanumeric characters display ended
		window.requestAnimationFrame(displayDelay);  // set another callback
	}
}


// handle delay period between end of alphanumeric characters and begin of arrows
// trigger begin of arrows display
function displayDelay(timestamp)  {
	let elapsed = timestamp - startTime; 
	if ( elapsed < ad_time )  {   // 	 var ad_time => arrow delay time (ms)
		window.requestAnimationFrame(displayDelay);  // wait
	}
	else  {   	
		document.getElementById('can_a').style.visibility = 'visible';  // only once  
		startTime = timestamp;	// timestamp when the arrows display start
		window.requestAnimationFrame(displayArrow);  // set another callback - last
	}
}


// handle arrows duration  
// trigger end of arrows display and handle after period => setTimeOut() for showItem() 
function displayArrow(timestamp)  {
	let elapsed = timestamp - startTime; 
	if ( elapsed < a_time )  {   // 	var a_time =>  Arrow duration time (ms)
		window.requestAnimationFrame(displayArrow);  // wait
	}
	else  {     	
		document.getElementById('can_a').style.visibility = 'hidden';  // only once	
		setTimeout(showItem, after);  // this time period is not precize
	}
}


// https://docs.nagra.com/docs/os/524/maximizing-ui-performance
// https://developer.mozilla.org/en-US/docs/Web/API/Canvas_API/Tutorial/Optimizing_canvas
// http://creativejs.com/resources/requestanimationframe/
// https://developer.mozilla.org/en-US/docs/Web/API/window/requestAnimationFrame
// https://stackoverflow.com/questions/30830126/what-is-difference-between-fillrect0-0-0-1-and-clearrect	
// browser calls this function to update an animation before the next repaint (callback)
// this is 1. callback. main function: prepare stimulus in period before stimulus
function draw(timestamp) 
{
		
/* var before = 1000;     //  Delay time before show each stimulus (ms).
   var s_time = 1000;     //  Alphanumeric characters duration (ms)
	 var ad_time = 1000;    //  Arrow delay time (ms)
	 var a_time = 1000;     //  Arrow duration time (ms)
	 var after = 1000;			//  Delay time after show each stimulus (ms)  */

	if (counter == 0)  {   		
		startTime = timestamp;	// reset time when the stimulus sequence started 
    drawString();			// prepare main stimulus
		drawArrow(document.getElementById('can_a'), true); // prepare arrows
		document.getElementById('can_s').style.visibility = 'hidden';
		document.getElementById('can_a').style.visibility = 'hidden';
	}	
	counter= 1;
	let elapsed = timestamp - startTime;  // const elapsed = timestamp - startTime; 					
	if ( elapsed <= before )  {  // wait => var before => Delay time before show each stimulus (ms)
		window.requestAnimationFrame(draw); // Note: requestAnimationFrame() is 1 shot.
	}
	else  { 		  
		startTime = timestamp;	// timestamp when the alphanumeric characters display started 
		document.getElementById('can_s').style.visibility = 'visible';  // only once
		window.requestAnimationFrame(displayString);  // set another callback => dispatch
	}		
}


// draw main stimulus (alphanumeric characters) 
function drawString() 
{
	let c = document.getElementById('can_s'), ctx = c.getContext('2d'), w = c.width, h = c.height, txt;					
	ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, w, h);    // ctx.clearRect(0, 0, w, h);
  ctx.fillStyle= pen_clr;
	let width = c.width / (alfa_cnt+2),  height = c.height / 3;  // cell width, cell height
	for (let row = 0; row <= 2; row++)
	{	   
		let cnt= 0;   // string in each cell array index
		for (let col= 1; col < alfa_cnt + 1; col++)
		{                						
			let left= col * width, top = row * height;  // cell left bound, cell top bound
			let xs = Math.round(left+width/2), ys= Math.floor(top+height/2);
			let r= Math.floor( Math.min(width,height)/2 );  
			ctx.font = r.toString() + 'px ' + font_family;   			      
			txt= stim[item][row][cnt];
			ctx.fillText(txt, left+width/2, top+height/2);
			cnt++;
		}
	}
  if ( is_arrow[item] == true && ad_time == 0 )  {   
    drawArrow(c, false);
	}		
}	


/*	draw arrows after main stimulus and parallel with main stimulus (if arrow delay time == 0)
		canvas => document.getElementById('can_a') or document.getElementById('can_s')
		is_clear => if true : repaint (use fillStyle), else only draw arrows */
function drawArrow(canvas, is_clear) 
{

	let ctx = canvas.getContext('2d'), w = canvas.width, h = canvas.height;					
	if ( is_clear )  {
		ctx.clearRect(0, 0, w, h);
		ctx.fillStyle= back_clr;	
	}    //ctx.fillStyle= pen_clr; see: arrow()
	
	if ( is_arrow[item] == true ) 
	{
		let is_left=  document.getElementById('arrows').value != '2'  ? true : false;  // '2' => right side
		let is_right= document.getElementById('arrows').value != '1'  ? true : false;  // '1' => left side

		let width = w / (alfa_cnt+2), height = h / 3;  // cell width, cell height
		let row = t_row[item];
		
		for (let i= 0; i<= 1; i++)	
		{
			let col, left, top, x0, y0, r; 	           						
			col =  i == 0 ? 0 : alfa_cnt + 1;
			left = col * width;   // cell left bound
			top = row * height;  // cell top bound
			x0 = Math.round(left+width/2);  // cell x-center
			y0= Math.floor(top+height/2);   // cell y-center
			r = Math.min(width,height)/2;  // cell radius
			r= Math.floor(r*0.5);		
			if (i == 0 && is_left) {
				arrow(ctx, x0, y0, r, pen_clr, 3);
			}
			if (i== 1 && is_right) {
				arrow(ctx, x0, y0, r, pen_clr, 2);
			}	
		}
	}

}	


/* draw Arrow
    ctx = canvas (object) context,
    x = x-center,
    y= y-center,
    r = radius, 
    col = color, 
    (par ===0) => vertical UP arrow
    (par===1)  =>  vertical DOWN arrow
    (par===2)  =>  horizontal LEFT arrow
    (par===3)  => horizontal RIGHT arrow 
*/
function arrow(ctx, x, y, r, col, par)
{    
  //if (par < 0 || par > 3) return;   
  var k1=0.25*r,    k2= 0.75*r; 
  // init path (points) for different par argument
  var p=  [  
  // vertical UP arrow => p[0]
  [  
     x-k1,y+r, 
     x+k1,y+r, 
     x+k1,y-k1, 
     x+k2,y-k1, 
     x,y-r, 
     x-k2,y-k1, 
     x-k1,y-k1     
   ],                                          
  //  vertical DOWN arrow => p[1]
  [ 
      x-k1,y-r, 
      x-k1,y+k1, 
      x-k2,y+k1,    
      x,y+r, 
      x+k2,y+k1,
      x+k1,y+k1,
      x+k1,y-r     
   ],    
  //  horizontal LEFT arrow => p[2]
  [
      x+r, y+k1, 
      x-k1,y+k1, 
      x-k1,y+k2 ,   
      x-r,y,
      x-k1,y-k2,
      x-k1,y-k1,
      x+r,y-k1        
  ],        
  //  horizontal RIGHT arrow => p[3]
  [
     x-r, y+k1, 
     x+k1,y+k1, 
     x+k1, y+k2,    
     x+r,y, 
     x+k1,y-k2,
     x+k1,y-k1,
     x-r,y-k1          
  ]  
  ];
   
  ctx.beginPath();
  ctx.fillStyle= col; 
  for (var j=0; j<p[par].length; j=j+2)  {
    (j==0) ? ctx.moveTo(p[par][j],  p[par][j+1]) : ctx.lineTo(p[par][j],  p[par][j+1]);            
  }    
  ctx.closePath();
  ctx.fill();        
}


// return random integer in range [min..max]
function getRndInt(min, max) {
		return Math.floor(Math.random() * (max - min + 1) ) + min;
}


// trigger delayed stimulus draw (animation frame)
function delayDraw()
{
	repeated[item]++;
	counter= 0;
	window.requestAnimationFrame(draw);	// activate callback Window.requestAnimationFrame()  	
}	


// show current item_div panel/div. items are in range: 0..item_cnt-1     
function showItem() 
{   		
  if ( repeated[item] == 0 )  {  // only first time when item is displayed	 
		document.getElementById('item_div').style.display= 'none';     
		setTimeout(delayDraw, before);
	} 
	else 
	{ 
		document.getElementById('item_div').style.display= 'block';  
		
		// Show the user the content of which row to enter (top, middle, bottom)
		// https://www.w3schools.com/charsets/ref_utf_arrows.asp
		for (let i= 0; i<= 2; i++) {
			document.getElementById('row_' + i).style.visibility = t_row[item] == i ? 'visible' : 'hidden';
		}
		if ( t_row[item] == 0 ) {
			document.getElementById('row_desc').textContent= 'top';
		}
		else if ( t_row[item] == 1 ) {
			document.getElementById('row_desc').textContent= 'middle';
		}
		else  {
			document.getElementById('row_desc').textContent= 'bottom';
		}
		
		// update footer (navigation buttons, progress bar,..)
		document.getElementById('prior').style.display= (item > 0 && is_back) ? 'inline': 'none';  // set footer navigation prior button (left arrow) 				 
		let btn= document.getElementById('next');     
		if(item == stim.length-1)  {       
			btn.style.backgroundImage=  "url('img/end.png')";  // update "Next item" image on "right arrow" button 
			btn.title= 'End of task';
		}
		else  {
			btn.style.backgroundImage= "url('img/forward.png')"; 
			btn.title= 'Next';
		} 
						
		if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       	
			let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= (item + 1) + ' / ' + item_cnt;   // item/total
		}
		
		document.getElementById('repeat').style.display= repeated[item] < rep_cnt ? 'block' : 'none';  // enable or disable stimulus repeating 
		
		// refresh string values for current item
		document.getElementById('answer').value= ans[item];	   // user answer
					
		document.getElementById('answer').focus();		
		
		if (demo)  {			
			let tmp= stim[item][t_row[item]];  // whole text in target row	
			if ( ttype == '3' )  {
				tmp= reverseString(tmp);
			}			
			document.getElementById('demo_sol').textContent= 'Target row= ' + tmp; 
		} 
		document.getElementById('req_val').style.display= 'none'; // hide required warning 
	}	   
}


// generate new stimulus within current item => repeat stimulus
// this event is trigger by user click
function repeatStim()
{
	t_row[item]= getRndInt(0,2);   // generate new target row
	newItem(item);
	repeat_count++;
	document.getElementById('item_div').style.display= 'none';
	document.getElementById('answer').value='';   // reset user answer 
	ans[item]= '';
	setTimeout(delayDraw, before);	
}


// end of task (research)
// if (is_error== true) => page (window) was resized during task (when is_task==true). Only show html error warning
function theEnd(is_error)  
{     
	is_task= false;    
	
	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
 
	if (is_error) {
		document.getElementById('error').style.display= 'block';
		document.getElementById('result_wrapper').style.display= 'none';
		document.getElementById('score').textContent= '';
		return;
	}
	
  document.getElementById('error').style.display= 'none';
	document.getElementById('result_wrapper').style.display= 'block';
	
	let mis= 0, sum= 0, is_mis, cor, point;   
	let rows = [], points= [], points_arrow=[], points_no_arrow=[];
  let header = ptype == '1' ? ['Item', 'Correct', 'Entered value', 'Points'] : ['Item', 'Correct', 'Entered value', 'Arrow', 'Points'];
 
	for (let i = 0; i < item_cnt; i++)  
	{
    is_mis = ans[i] == '' ? true : false;;
		cor = stim[i][t_row[i]];  // whole correct text in target row => correct answer as array
		let cor_s =  ttype == '3' ? reverseString(cor) : cor;
		if ( is_mis )  {  
		  mis++;		// mis = number of missing items in task 
			if ( ptype == '1' )  { 
				rows[i]=  [ parseInt(i+1,10), cor_s, '',  0 ];
      }
      else 	{
				rows[i]= is_arrow[i] == true ? [ parseInt(i+1,10), cor_s, '', 'Yes', 0 ] : [ parseInt(i+1,10), cor_s, '', 'No', 0 ];
			}
    }		
		else  
		{  	  
			if ( alfa_type == '2' ) {  		//  <option value="2">Letters-lowercase</option>
					ans[i]= ans[i].toLocaleLowerCase();
			}
			else if ( alfa_type == '3' )  {  // <option value="3">Letters-UPPERCASE</option>
			    ans[i]= ans[i].toLocaleUpperCase();
			}
			point= getScore(ans[i], cor);
      points.push(point);
      sum+= point;		// sum of points	
			if ( ptype == '1') {
				rows[i]= [ parseInt(i+1,10), cor_s, ans[i], point ]; 
			}
			else  {
				if ( is_arrow[i] == true )  {
					rows[i]=  [ parseInt(i+1,10), cor_s, ans[i], 'Yes', point ];
				  points_arrow.push(point);
				}
        else {
				  rows[i]=	[ parseInt(i+1,10), cor_s, ans[i], 'No', point ]; 
				  points_no_arrow.push(point);
				}
			}
		}					
	}		
  if ( ptype == '1' ) {
		rows.push( ['Sum', '', '', sum ] );	
	}
	else  {
		rows.push( ['Sum', '', '', '', sum ] );
	}
	createTableArray(document.getElementById('table1'), rows, header);	
	
	// calculate decriptive statistics for the points => table2  	  
	const D = 3;  // number of decimals
	rows = [];   // reset table rows (for table2)		 
	header=  ['Group/Variable', 'Average', 'Standard Deviation', 'N']; // header
	
	let m1= '', sd1= '', m2= '', sd2= '';
	if ( ptype == '2')  {
		let n1 = points_arrow.length; 
		m1 = ( n1 > 0 ) ? mean(points_arrow).toFixed(D) : '';      // mean only for non-missing items in task WITH arrows in delay period
		sd1 = ( n1 > 1 ) ? stDev(points_arrow, m1).toFixed(D): '';  // standard deviation only for non-missing items in task WITH arrows in delay period 
		rows.push( ['Points with Arrow', m1, sd1, n1] );  
		let n2 = points_no_arrow.length;
		m2 = ( n2 > 0 ) ? mean(points_no_arrow).toFixed(D) : '';      // mean only for non-missing items in task with NO arrows in delay period
		sd2 = ( n2 > 1 ) ? stDev(points_no_arrow, m2).toFixed(D): '';  // standard deviation only for non-missing items in task with NO arrows in delay period 
	  rows.push( ['Points without Arrow', m2, sd2, n2] );  
	}
	let n =  points.length;  // number of all non-missing items
	let m = ( n > 0 ) ? mean(points).toFixed(D) : '';      // mean only for total non-missing items in task
	let sd = ( n > 1 ) ? stDev(points, m).toFixed(D): '';  // standard deviation only for total non-missing items in task 
	rows.push( ['Points all items', m, sd, n] );  // last row
	
	createTableArray(document.getElementById('table2'), rows, header);  
		
	// update 'score' label => above "Repeat" button
	document.getElementById('score').textContent= 'Sum of points= ' + sum + '. '; 
	switch (n)  {
		case 0: 
			document.getElementById('score').textContent+= 'Descriptive statistics cannot be calculated. Missing items= ' + mis;
			break;
		case 1: 
			document.getElementById('score').textContent+= 'Average points= ' + m + '  Standard deviation cannot be calculated. Missing items= ' + mis;
			break;	
		default:
			document.getElementById('score').textContent+= 'Average points= ' + m + '  Standard Deviation= ' + sd + '. Missing items= ' + mis; 
	}
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
		
	try 
  {
    if (store) {
			setStorage(sum, mis, m, sd, m1, sd1, m2, sd2);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
 
}


/* user => user entered value for this item. 
   correct => correct answer (whole target row content) for this item => join as string!
   this function is called for each item
   return number of points => For each stimulus: 1 point is added for each correctly entered alphanumeric character within the target row  */
function getScore(user, correct)
{
	user= trim(user);  // redundant
	let res= 0, up= correct.length;
	if ( ttype == '3' )  {
		correct= reverseString(correct);
	}			
	if ( ttype == '1')  {
		for (let i= 0; i < up; i++)  { 
			if ( user.indexOf(correct[i]) != -1 ) {     //  str.indexOf(find, str.length) !== -1;
				res++;
			}
		}
	}
	else 	{  // left-to-right or right-to-left
		for (let i= 0; i < up; i++)  {
			if ( typeof user[i] === 'undefined' ) {
				continue;  // break; 
			}
			else if ( user[i] == correct[i] ) {
				res++;  
			}
		}			
	}
	return res;
}


// https://www.freecodecamp.org/news/how-to-reverse-a-string-in-javascript-in-3-different-ways-75e4763c68cb/
function reverseString(str) {
  return str.split("").reverse().join("");
}


// return arithmetic mean (mean, average) of input array (arr)
// no validation => input array must be in valid number format	
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction) => N-1
// no validation => input array must be in valid number format
// arr = array argument;  mean = average for input arr 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
		sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}

 
/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
	 sum = sum of points in task
	 mis = number of missing items in task
	 m = mean only for non-missing items in task
   sd = standard deviation only for non-missing items in task  
	 m1 = mean only for non-missing items WITH arrow in delay period
	 sd1 = standard deviation only for non-missing items WITH arrow in delay period
	 m2 = mean only for non-missing items WITHOUT arrow in delay period
	 sd2 = standard deviation only for non-missing items WITHOUT arrow in delay period  */
function setStorage(sum, mis, m, sd, m1, sd1, m2, sd2)   
{   
	let oldItems = JSON.parse(localStorage.getItem('prt-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		        
			'points': sum, 
			'missing' : mis,
			'mean' : m,
			'sd' : sd,	
      'mean-arrow' : m1,
			'sd-arrow' : sd1,
			'mean-no-arrow' : m2,
			'sd-no-arrow' : sd2,			
			'paradigm' : ptype,
			'items': item_cnt,
			'alpha-type': alfa_type,
			'alpha-count' : alfa_cnt,
			'task-type' : ttype,
			'arrow-pos' : arrows,
			'repeat-max': rep_cnt,	
			'repeat-count': repeat_count,
			'before-time': before,
			'main-stimulus' : s_time,
			'arrow-delay ': ad_time,		
			'arrow-time': a_time,		
			'after-time' : after,			
			'backward': is_back ? 1 : 0,
			'is-missing' : is_mis ? 1 : 0,  			      
			'demo': demo ? 1 : 0		
  };

  oldItems.push(newItem);

  localStorage.setItem('prt-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);  
 
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined)	{
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').setAttribute('class', 'center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}

// begin of new code for export localStorage app data:  data= localStorage.getItem('myApp-v1.0');

// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (document.getElementById('store_group').disabled === false && localStorage.getItem('prt-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('prt-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message	
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('prt-v1.0') ? JSON.parse(localStorage.getItem('prt-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application results: '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('prt-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('prt-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('prt-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// displays Latest research results in new empty web page (tab) 
function newWinLast()
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  // note: id='table2' is html div element
			break;
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
